#!/usr/bin/env python
# encoding: utf-8

import time
START_TIME = time.time()

import os, sys
topdir = os.path.abspath(os.path.dirname(__file__))

# gettext installation
if os.name == 'nt':
   # windows hack for locale setting
   import locale
   lang = os.getenv('LANG')
   if lang is None:
      default_lang, default_enc = locale.getdefaultlocale()
      if default_lang:
         lang = default_lang
      if lang:
         os.environ['LANG'] = lang
import gettext
gettext.install("dsgui", os.path.join(topdir, "locale"))

# import and check gtk
GTK_REQUIRE = "2.0"
try:
  import pygtk
  pygtk.require("2.0") 
except:
  print _("Required PyGTK 2.0 was not found - try installing python-gtk, pygtk or similar package.")
  sys.exit()
try:
  import gtk
except:
  print _("Python module 'gtk' was not found - try installing python-gtk, pygtk or similar package.")

if __name__ == "__main__":
  GTK_REQUIRED_MINIMUM = (2,14,0)
  check = gtk.check_version(*GTK_REQUIRED_MINIMUM)
  if check: # check did not pass
    current_version = gtk.gtk_version
    message = _("""Your version of GTK is older that the required minimum.
  The application will continue, but it will probably not work properly.
  
  Information about versions:
    Your version: %d.%d.%d
    Required version: %d.%d.%d""") % (current_version+GTK_REQUIRED_MINIMUM)
    d = gtk.MessageDialog(None, message_format=message, buttons=gtk.BUTTONS_OK, type=gtk.MESSAGE_WARNING)
    d.run()
    d.destroy()
  
try:
  import gnomevfs
  GNOMEVFS_AVAILABLE = True
except:
  GNOMEVFS_AVAILABLE = False
import pango

from dslib.client import Client
from ds_search import DsSearchDialog
from ds_send import DsSendDialog
from proxy_dialog import ProxyDialog
from xml_translator import translate_xml_string


class GUI(object):

  topdir = topdir
  date_time_format = "%Y-%m-%d %H:%M"
  account_info_attrs = (("dbID",_("Data box ID")),("dbType",_("Data box type")),
                      ("ic",_("IČ")),("pnFirstName",_("Given name")),
                      ("pnMiddleName",_("Middle name")),("pnLastName",_("Surname")),
                      ("pnLastNameAtBirth",_("Surname at birth")),("firmName",_("Firm name")),
                      ("biDate",_("Date of birth")),("biCity",_("City of birth")),
                      ("biCounty",_("County of birth")),("biState",_("State of birth")),
                      ("adCity",_("City of residence")),("adStreet",_("Street of residence")),
                      ("adNumberInStreet",_("Number in street")),("adNumberInMunicipality",_("Number in municipality")),
                      ("adZipCode",_("Zip code")),("adState",_("State of residence")),
                      ("nationality",_("Nationality")),("dbEffectiveOVM",_("Effective OVM")),
                      ("dbOpenAddressing",_("Open addressing")),
                      )

  def __init__(self):
    self.proxy = -1
    self.username = None
    self.password = None
    self.test_account = None
    self.received_messages = {} # id2message
    self.sent_messages = {}
    self.message_data = {} # id to data
    self.account_info = None
    # build the GUI from xml
    builder = gtk.Builder()
    #builder.set_translation_domain("dsgui")
    self.builder = builder
    gui_xml = GUI.fix_ui_xml("dsgui.ui")
    if gtk.gtk_version < (2,16,0):
      from ui_xml_versions import remove_incompatible_parts
      gui_xml_optimized = remove_incompatible_parts(gui_xml, gtk.gtk_version)
    else:
      gui_xml_optimized = gui_xml
    try:
      builder.add_from_string(gui_xml_optimized.decode('utf-8'))
    except:
      # this fixes problem with different API on windows and PyGTK 2.12
      builder.add_from_string(gui_xml_optimized.decode('utf-8'), -1)
    self.window1 = builder.get_object("window1")
    self._post_builder_ui_compatibility_fixes(gui_xml)
    builder.connect_signals(self)
    self.window1.show()
    tag_table = builder.get_object("message_tagtable")
    self.received_message_store = builder.get_object("received_message_store")
    self.received_message_treeview = builder.get_object("received_message_treeview")
    self.received_message_textview = builder.get_object("received_message_textview")
    self.received_attachment_treeview = builder.get_object("received_attachment_treeview")
    self.received_attachment_store = builder.get_object("received_attachment_store")
    self.sent_message_store = builder.get_object("sent_message_store")
    self.sent_message_treeview = builder.get_object("sent_message_treeview")
    self.sent_message_textview = builder.get_object("sent_message_textview")
    self.account_store = builder.get_object("account_store")
    self.account_treeview = builder.get_object("account_treeview")
    self.statusbar = builder.get_object("statusbar")
    self._download_context = self.statusbar.get_context_id("download")
    # the following should not be necessary, but Glade somehow does not allow
    # to populate a tagtable with tags
    # UPDATE - it is done in a different way below
    #tag_table = builder.get_object("message_tagtable")
    #tag_table.add(builder.get_object("bold_texttag"))
    #tag_table.add(builder.get_object("larger_texttag"))
    # for some reason, "message_tagtable" is associated only with the "sent" buffer,
    # not with "received" - I fix it here by hand
    # UPDATE - text tags don't work well with glade anyway - lets do it all by hand
    for tv_id in ("sent_message_textview","received_message_textview","account_info_textview"):
      tb = self.builder.get_object(tv_id).get_buffer()
      try:
        tb.create_tag("bold", weight=pango.WEIGHT_BOLD)
      except:
        pass
      try:
        tb.create_tag("larger", scale=1.25, pixels_above_lines=10)
      except:
        pass
    # some tweaks to permit nice display of read messages
    col = builder.get_object("received_message_read_column")
    cell = builder.get_object("received_message_read_cell_renderer")
    col.set_cell_data_func(cell, self._render_read_icon)
    # bind some special features to username rendering
    col = builder.get_object("account_treeview_username_column")
    cell = builder.get_object("account_treeview_username_cellrenderer")
    col.set_cell_data_func(cell, self._render_account_name)
    # add sorting info by hand - it does not load properly from XML
    # in older version of gtk (2.16 and below)
    for colname, i in (("treeviewcolumn_id", 0),("treeviewcolumn_title",1),
                       ("treeviewcolumn_sender",2), ("treeviewcolumn7",3),
                       ("treeviewcolumn8",4), ("received_message_read_column",5),
                       ("treeviewcolumn3",0), ("treeviewcolumn4",1),
                       ("treeviewcolumn5",2), ("treeviewcolumn6",6),
                       ("treeviewcolumn9",4), ("treeviewcolumn10",5),
                       ):
      col = builder.get_object(colname)
      col.set_sort_column_id(i)
    # bind text boldness to read locally status for received_attachment_treeview
    for colname in ("treeviewcolumn_id","treeviewcolumn_title","treeviewcolumn_sender",
                    "treeviewcolumn7","treeviewcolumn8"):
      col = builder.get_object(colname)
      renderer = col.get_cell_renderers()[0]
      col.set_cell_data_func(renderer, self._render_read_text)
    # check if python version works with HTTPS proxy
    python_version = tuple(sys.version_info[0:3])
    if python_version >= (2,6,3):
      self.proxy_support_ok = True
    else:
      self.proxy_support_ok = False
    if not self.proxy_support_ok:
      action = self.builder.get_object("change_proxy_settings")
      action.set_sensitive(False)
      try:
        # set_tooltip is not supported in older versions of PyGTK
        action.set_tooltip(_("Your version of Python does not support HTTPS proxy!"))
      except:
        pass
      self.proxy = None
      self.statusbar.push(-1, _("Your version of Python does not support HTTPS proxy!"))
    # activate tooltip display in statusbar for menu
    self._activate_tooltips_for_main_menu()
    # load a stored config file
    self._load_config_file()
    # open an account
    if not self.username:
      self.on_add_account_activate()
    self._check_empty_message_stores()
    #print time.time() - START_TIME

      
  # Callbacks
  
  def on_connect_action_activate(self, widget):
    received_count, sent_count = 0, 0
    # get received messages
    reply, error = self.try_client_method(
      self.client.GetListOfReceivedMessages, (),
      error_message=_("It was not possible to download the list of received messages.")
      )
    if not error:
      if reply.status.dmStatusCode == "0000":
        for m in reply.data:
          self.add_received_message(m)
          received_count += 1
      # get sent messages
      reply, error = self.try_client_method(
        self.client.GetListOfSentMessages, (),
        error_message=_("It was not possible to download the list of sent messages.")
        )
      if not error:
        if reply.status.dmStatusCode == "0000":
          for m in reply.data:
            self.add_sent_message(m)
            sent_count += 1
    self.statusbar.push(self._download_context,
                        _("%d received and %d sent messages found on server")%(received_count,sent_count))
    self._check_empty_message_stores()
    if not error:
      # check info about the account
      reply, error = self.try_client_method(
        self.client.GetOwnerInfoFromLogin,
        (),
        error_message=_("It was not possible to download info about the account."))
      if not error:
        self.set_account_info(reply.data)


  def on_about_activate(self, widget):
    dial = self.builder.get_object("aboutdialog1")
    dial.run()
    dial.hide()

  def on_received_message_treeview_cursor_changed(self, treeview):
    message = self._get_selected_received_message()
    if message:
      self.show_received_message(message)
      # scroll to selection - it can disappear when sorting by read locally status
      # and reading something
      model, paths = self.received_message_treeview.get_selection().get_selected_rows()
      if len(paths) == 1:
        self.received_message_treeview.scroll_to_cell(paths[0])
    self.on_received_attachment_treeview_cursor_changed(treeview)


  def on_window1_destroy(self, w):
    self._save_state()
    gtk.main_quit()


  def on_download_complete_message_activate(self, w):
    message = self._get_selected_received_message()
    reply, error = self.try_client_method(
      self.client.MessageDownload,
      (message.dmID,),
      error_message=_("It was not possible to download the message."),
      )
    if not error:
      message = reply.data
      self.add_received_message(message)
      self.show_received_message(message)
      self.on_received_attachment_treeview_cursor_changed(None)

  def on_received_attachment_treeview_row_activated(self, tree, path, column):
    self.builder.get_object("open_attachment").activate()

  def on_save_attachment_activate(self, w):
    message, attachment = self._get_selected_attachment()
    if attachment:
      d = gtk.FileChooserDialog(title=_("Select where to save"),
                                parent=self.window1,
                                action=gtk.FILE_CHOOSER_ACTION_SAVE,
                                buttons=(gtk.STOCK_OK,gtk.RESPONSE_ACCEPT,
                                         gtk.STOCK_CANCEL,gtk.RESPONSE_REJECT)
                                )
      d.set_current_name(attachment._dmFileDescr)
      ok = d.run()
      if ok == gtk.RESPONSE_ACCEPT:
        fullname = d.get_filename()
        dir, name = os.path.split(fullname)
        attachment.save_file(dir, name)
      else:
        pass
      d.destroy()

  def on_open_attachment_activate(self, w):
    message, attachment = self._get_selected_attachment()
    done = False
    if GNOMEVFS_AVAILABLE:
      mime_type = gnomevfs.get_mime_type_for_data(attachment.get_decoded_content())
      mime_app = gnomevfs.mime_get_default_application(mime_type)
      if mime_app:
        temp_file = self._create_temp_file_for_dmfile(attachment)
        import subprocess
        application = subprocess.Popen([mime_app[2],temp_file])
        done = True
    if not done:
      d = gtk.MessageDialog(self.window1,
                            message_format="I could not find an application to open this attachment. You can save it and open it manually.",
                            buttons=gtk.BUTTONS_OK,
                            type=gtk.MESSAGE_ERROR)
      d.run()
      d.destroy()
      

  def on_sent_message_treeview_cursor_changed(self, treeview):
    message = self._get_selected_sent_message()
    self.show_sent_message(message)

    
  def on_edit_credentials_activate(self, w):
    d = self.builder.get_object("credentials_dialog")
    login_entry = self.builder.get_object("login_entry")
    login_entry.set_text(self.username or "")
    password_entry = self.builder.get_object("password_entry")
    password_entry.set_text(self.password or "")
    test_account_button = self.builder.get_object("test_environment_checkbutton")
    if self.test_account == None:
      # make test account the default 
      test_account_button.set_active(True)
    else:
      test_account_button.set_active(self.test_account)
    result = d.run()
    if result == 1:
      username = login_entry.get_text()
      password = password_entry.get_text()
      test_account = test_account_button.get_active()
      for row in self.account_store:
        if row[0]==self.username and row[2]==self.test_account:
          row[0] = username
          row[1] = password
          row[2] = test_account
          break
      self._open_known_account(username)
    else:
      pass # cancel or something like this was pressed
    d.hide() # hide, don't destroy - the dialog is created by Builder


  def on_notebook1_switch_page( self, w, page, page_num):
    if page_num != 0:
      self.builder.get_object("download_complete_message").set_sensitive(False)
      self.builder.get_object("save_attachment").set_sensitive(False)
      self.builder.get_object("open_attachment").set_sensitive(False)
    else:
      self.on_received_message_treeview_cursor_changed(None)


  def on_find_data_box_activate(self, w):
    d = DsSearchDialog(self, self.client)
    result = d.run()
    d.hide()
    if result == 1:
      boxes = d.get_selected_box_names()
      m = "\n".join(["%s: %s"%(box[0],box[1]) for box in boxes])
      d = gtk.MessageDialog(self.window1,
                            message_format=_("You selected the following boxes:\n\n")+m,
                            buttons=gtk.BUTTONS_OK,
                            type=gtk.MESSAGE_INFO)
      d.run()
      d.destroy()
    
  def on_send_message_activate(self, w):
    dial = DsSendDialog(self, self.client)
    ok = False
    new = True
    while not ok:
      if not new:
        dial.show_all()
      new = False
      result = dial.run()
      dial.hide()
      if result == 1:
        # dummy operation at first to get a cookie
        reply, error = self.try_client_method(self.client.DummyOperation, (), handle_status=False)
        if not error:
          # now send the file
          error_message = _("It was not possible to send the message")
          envelope, dmfiles = dial.get_prepared_message()
          reply, error = self.try_client_method(self.client.CreateMessage,
                                                 (envelope, dmfiles),
                                                 error_message=error_message,
                                                 handle_status=False)
          message = reply.status.dmStatusMessage
          code = reply.status.dmStatusCode
          if code != "0000":
            mtype = gtk.MESSAGE_ERROR
            message_text = _("An error occured when sending the message:")
          else:
            mtype = gtk.MESSAGE_INFO
            message_text = _("Message was successfully sent:")
            ok = True
          message_text += "\n\n"
          message_text += _("Reply status: %(message)s\nReply code: %(code)s")%{"message":message,"code":code}
          d = gtk.MessageDialog(self.window1,
                                message_format=message_text,
                                buttons=gtk.BUTTONS_OK,
                                type=mtype)
          d.run()
          d.destroy()
      else:
        break

  def on_received_message_treeview_row_activated(self, tree, path, column):
    read_col = self.builder.get_object("received_message_read_column")
    if column == read_col:
      model = tree.get_model() 
      iter = model.get_iter(path) 
      msgid = model.get_value(iter, 0)
      read = self.message_data[msgid]['read_locally']
      read = not read
      self.message_data[msgid]['read_locally'] = read
      model.set_value(iter, 5, read)
    else:
      # open the message in a new window
      dial = self.builder.get_object("show_message_dialog")
      message = self._get_selected_received_message()
      textview = self.builder.get_object("show_message_dialog_textview")
      self.show_received_message(message, textview=textview)
      dial.show_all()
      dial.run()
      dial.hide()

  def on_sent_message_treeview_row_activated(self, tree, path, column):
    # open the message in a new window
    dial = self.builder.get_object("show_message_dialog")
    message = self._get_selected_sent_message()
    textview = self.builder.get_object("show_message_dialog_textview")
    self.show_sent_message(message, textview=textview)
    dial.show_all()
    dial.run()
    dial.hide()

  def _something_with_attachment_happened(self):
    message, attachment = self._get_selected_attachment()
    if attachment:
      self.builder.get_object("save_attachment").set_sensitive(True)
      self.builder.get_object("open_attachment").set_sensitive(True)
    else:
      self.builder.get_object("save_attachment").set_sensitive(False)
      self.builder.get_object("open_attachment").set_sensitive(False)
    
      
  def on_received_attachment_store_row_number_changed(self, model, path, iter=None):
    self._something_with_attachment_happened()

  def on_received_attachment_treeview_cursor_changed(self, treeview):
    self._something_with_attachment_happened()

  def on_account_treeview_cursor_changed(self, treeview):
    account = self._get_selected_account()
    if account:
      account = list(account)[:3]
      self._open_account(*account)
      
  def on_add_account_activate(self, w=None):
    d = self.builder.get_object("credentials_dialog")
    login_entry = self.builder.get_object("login_entry")
    login_entry.set_text("")
    password_entry = self.builder.get_object("password_entry")
    password_entry.set_text("")
    test_account_button = self.builder.get_object("test_environment_checkbutton")
    test_account_button.set_active(True)
    result = d.run()
    if result == 1:
      username = login_entry.get_text()
      password = password_entry.get_text()
      test_account = test_account_button.get_active()
      row_id = self.add_account(username, password, test_account)
      if row_id >= 0:
        self.account_treeview.set_cursor(row_id)
      else:
        message = _("%(account_type)s '%(username)s' is already present, you cannot add it twice!")
        account_type = self._get_account_type_as_text(test_account)
        message = message % locals()
        dial = gtk.MessageDialog(self.window1,
                                 message_format=message,
                                 buttons=gtk.BUTTONS_OK,
                                 type=gtk.MESSAGE_ERROR)
        dial.run()
        dial.destroy()
    else:
      # cancel or something like this was pressed
      self._check_zero_accounts()
    d.hide() # hide, don't destroy - the dialog is created by Builder

  def on_account_treeview_key_release_event(self, w, event):
    if event.keyval == gtk.keysyms.Delete:
      account = self._get_selected_account()
      if account:
        d = gtk.MessageDialog(self.window1,
                              message_format=_("Remove account '%s'") % account[0],
                              buttons=gtk.BUTTONS_OK_CANCEL,
                              type=gtk.MESSAGE_QUESTION)
        answer = d.run()
        if answer == gtk.RESPONSE_OK:
          treeview = self.account_treeview
          model, paths = treeview.get_selection().get_selected_rows()
          self.account_store.remove(self.account_store.get_iter(paths[0]))
        d.destroy()
        
  def on_show_help_index_activate(self, w=None):
    import webbrowser
    webbrowser.open(GUI.topdir+"/doc/index.html")
    url = "file://"+GUI.topdir+"/doc/index.html"
    d = gtk.MessageDialog(self.window1,
                          message_format=_("Help index should open in your browser.\nIf it does not, you can use the following URL:\n%s")%url,
                          buttons=gtk.BUTTONS_OK,
                          type=gtk.MESSAGE_INFO)
    d.run()
    d.destroy()
    
  def on_change_proxy_settings_activate(self, w):
    dial = ProxyDialog(self, self.client)
    result = dial.run()
    if result == 1:
      proxy = dial.get_proxy()
      self.set_proxy(proxy)
    dial.destroy()

#  def on_received_message_store_row_deleted(self, model, path):
#    self._check_empty_message_stores()
#    
#  def on_received_message_store_row_inserted(self, model, path, iter):
#    self._check_empty_message_stores() 
#  
#  def on_sent_message_store_row_deleted(self, model, path):
#    self._check_empty_message_stores()
#    
#  def on_sent_message_store_row_inserted(self, model, path, iter):
#    self._check_empty_message_stores()   

  # // end of Callbacks

  # helper methods for display

  def _render_read_icon(self, column, cell, model, iter):
    data = model.get_value(iter, 5)
    if data == True:
      stock = gtk.STOCK_YES
    else:
      stock = gtk.STOCK_NO
    pixbuf = self.window1.render_icon(stock_id=stock, size=gtk.ICON_SIZE_MENU)
    cell.set_property('pixbuf', pixbuf)
    
  def _render_account_name(self, column, cell, model, iter):
    data = model.get_value(iter, 2)
    if data == True:
      style = pango.STYLE_ITALIC
    else:
      style = pango.STYLE_NORMAL
    cell.set_property('style', style)

  def _render_read_text(self, column, cell, model, iter):
    data = model.get_value(iter, 5)
    if data == True:
      cell.set_property('weight', pango.WEIGHT_NORMAL)
    else: 
      cell.set_property('weight', pango.WEIGHT_BOLD)


  def _fill_message_data_into_textbuffer(self, message, textbuffer, additional_lines=None):
    def add_row(heading, text, add_tags=None):
      add = add_tags or ()
      position = textbuffer.get_end_iter()
      textbuffer.insert_with_tags_by_name(position, heading, "bold", *add)
      position = textbuffer.get_end_iter()
      textbuffer.insert_with_tags_by_name(position, " %s\n"%text, *add)
    def add_newline():
      position = textbuffer.get_end_iter()
      textbuffer.insert_with_tags_by_name(position, "\n")  

    # fill the text buffer
    textbuffer.set_text("")
    add_row(_("Subject:"), message.dmAnnotation, add_tags=("larger",))
    for name, value in (("dmLegalTitleLaw",_("Law")),
                        ("dmLegalTitleYear",_("Year")),
                        ("dmLegalTitleSect",_("Section")),
                        ("dmLegalTitlePar",_("Paragraph")),
                        ("dmLegalTitlePoint",_("Letter"))):
      if getattr(message, name):
        add_row(value+":", getattr(message, name))
    add_newline()
    add_row(_("From:"), message.dmSender)
    add_row(_("Sender Address:"), message.dmSenderAddress)
    add_newline()
    add_row(_("To:"), message.dmRecipient)
    add_row(_("Recipient Address:"), message.dmRecipientAddress)
    add_newline()
    add_row(_("Delivery time:"), self._format_datetime(message.dmDeliveryTime))
    add_row(_("Acceptance time:"), self._format_datetime(message.dmAcceptanceTime))
    add_newline()
    add_row(_("Status:"), "%d - %s" % (message.dmMessageStatus, message.get_status_description()))
    add_newline()
    if additional_lines:
      for head, text in additional_lines:
        add_row(head, text)
    # events - we would need to get delivery info for this, leave it for now
    #add_newline()
    #add_row("", "---------------------------------------")
    #for event in message.dmEvents:
    #  add_row(event.dmEventTime, event.dmEventDescr)
    # signatures - once download of signed messages works, info will go here
    
    

  # // end of helper methods for display


  def show_received_message(self, message, textview=None):
    if message:
      if not textview:
        textview = self.received_message_textview
      attachment_store = self.received_attachment_store
      textbuffer = textview.get_buffer()
      if hasattr(message, 'dmFiles') and message.dmFiles:
        attachments = (_("Attachments:"), _("%d (downloaded and ready)")%len(message.dmFiles))
        self.builder.get_object("download_complete_message").set_sensitive(False)
      else:
        attachments = (_("Attachments:"), _("not downloaded yet, ~%s KB; use 'Download' to get them.")%message.dmAttachmentSize)
        self.builder.get_object("download_complete_message").set_sensitive(True)

      self._fill_message_data_into_textbuffer(message, textbuffer, additional_lines=[attachments])
      # show the attachments
      attachment_store.clear()
      for i,f in enumerate(message.dmFiles):
        attachment_store.append([i,f._dmFileDescr,f.get_size(),f._dmMimeType,f._dmFormat])
      # mark message read
      mid = int(message.dmID)
      if mid in self.message_data:
        if not self.message_data[mid]['read_locally']:
          self.message_data[mid]['read_locally'] = True
          for row in self.received_message_store:
            if row[0] == mid:
              row[5] = True
              break


  def show_sent_message(self, message, textview=None):
    if message:
      if not textview:
        textview = self.sent_message_textview
      textbuffer = textview.get_buffer()
      textbuffer.set_text("")
      self._fill_message_data_into_textbuffer(message, textbuffer)
      

  def add_received_message(self, message):
    msgid = int(message.dmID)
    if msgid not in self.message_data:
      self.message_data[msgid] = {'read_locally': False}
    if msgid in self.received_messages:
      if self.received_messages[msgid].get_origin() == "tReturnedMessage":
        # this is a complete message - do not replace it
        # (it could be useful to create a list of increasing origin importance)
        pass
      else:
        self.received_messages[msgid] = message
    else:
      self.received_messages[msgid] = message
      delivery_time = self._format_datetime(message.dmDeliveryTime)
      accept_time = self._format_datetime(message.dmAcceptanceTime)
      read = self.message_data[msgid]['read_locally']
      data = [msgid,message.dmAnnotation,message.dmSender,delivery_time,accept_time,read]
      self.received_message_store.append(data)


  def add_sent_message(self, message):
    msgid = int(message.dmID)
    self.sent_messages[msgid] = message
    delivery_time = self._format_datetime(message.dmDeliveryTime)
    accept_time = self._format_datetime(message.dmAcceptanceTime)
    status = message.dmMessageStatus
    status_str = message.get_status_description()
    data = [msgid,message.dmAnnotation,message.dmRecipient,status_str,delivery_time,accept_time,status]
    for i,row in enumerate(self.sent_message_store):
      if row[0] == msgid:
        self.sent_message_store[i] = data
        break
    else:
      self.sent_message_store.append(data)
      
  def add_account(self, username, password, test_account):
    for row in self.account_store:
      if row[0] == username and row[2] == test_account:
        # this account is already present - cannot add it twice
        return -1
    # try to add into to the tooltip of an account
    text = self._get_tooltip_for_account(username, test_account)
    iter = self.account_store.append([username,password,test_account,text])
    return self.account_store.get_path(iter)

  def set_account_info(self, info):
    self.account_info = info
    self._show_account_info()
    # update info in tooltip
    account = self._get_selected_account()
    if account:
      account[3] = self._get_tooltip_for_account(self.username, self.test_account)
      
  def set_proxy(self, proxy):
    if proxy != self.proxy:
      self.proxy = proxy
      # update client if proxy changed
      self.client = self._create_ds_client(self.username, self.password, self.test_account, proxy=self.proxy)  
      
  def _get_info_for_account(self, username, test_account):
    """get account info for an account identified by username and test_account"""
    # if current account is asked and self.account_info is available, use it 
    if username == self.username and self.test_account == test_account and \
      self.account_info:
      return self.account_info
    # otherwise get the data from the shelf
    key = GUI._create_shelf_key(username, test_account)
    shelf = self._open_shelf()
    data = shelf.get(key, None)
    info = None
    if data:
      info = data.get('account_info', None)
    shelf.close()
    return info
  
  def _get_tooltip_for_account(self, username, test_account):
    info = self._get_info_for_account(username, test_account)
    text = ""
    if info:
      for attr,name in GUI.account_info_attrs:
        if attr in ('dbID','pnFirstName','pnLastName',"biDate","firmName"):
          value = getattr(info, attr)
          if value:
            text += "<b>%s</b>: %s\n" % (name, value)
    text += "<b>%s</b>: %s" % (_("Account type"), self._get_account_type_as_text(test_account))
    return text

  def _show_account_info(self):
    # add info to the account_textview
    textbuffer = self.builder.get_object("account_info_textbuffer")
    textbuffer.set_text("")
    textbuffer.insert_with_tags_by_name(
      textbuffer.get_end_iter(),
      "%s\n" % self._get_account_type_as_text(self.test_account),
      'larger')
    textbuffer.insert_at_cursor("\n")
    info = self.account_info
    if info:
      for attr,name in GUI.account_info_attrs:
        value = getattr(info, attr)
        if value:
          position = textbuffer.get_end_iter()
          textbuffer.insert_with_tags_by_name(position, name+":", "bold")
          position = textbuffer.get_end_iter()
          textbuffer.insert_at_cursor(" %s\n"%value)
    else:
      position = textbuffer.get_end_iter()
      textbuffer.insert_at_cursor(_("Data has not been downloaded yet."))
   

  def _get_selected_received_message(self):
    treeview = self.received_message_treeview
    model, paths = treeview.get_selection().get_selected_rows()
    if len(paths) == 1:
      message = self.received_messages[model[paths[0]][0]]
      return message
    return None
    
  def _get_selected_attachment(self):
    message = self._get_selected_received_message()
    if message:
      treeview = self.received_attachment_treeview
      model, paths = treeview.get_selection().get_selected_rows()
      if len(paths) == 1:
        attachment = message.dmFiles[model[paths[0]][0]]
        return message, attachment
      return message, None
    return None, None

  def _get_selected_sent_message(self):
    treeview = self.sent_message_treeview
    model, paths = treeview.get_selection().get_selected_rows()
    if len(paths) == 1:
      message = self.sent_messages[model[paths[0]][0]]
      return message
    return None
      
      
  def _get_selected_account(self):
    treeview = self.account_treeview
    model, paths = treeview.get_selection().get_selected_rows()
    if len(paths) == 1:
      return model[paths[0]]
    return None

  def _save_state(self):
    # dump messages
    shelf = self._open_shelf()
    self._store_to_shelf(shelf)
    shelf.close()
    # save config file
    import ConfigParser
    config = ConfigParser.RawConfigParser()
    # call individual saving methods
    self._save_credentials(config)
    self._save_gui_state(config)
    self._save_connection_state(config)
    # write it out and chmod to user-only access
    cr_filename = self._get_config_filename()
    cr_file = file(cr_filename, "wb")
    config.write(cr_file)
    cr_file.close()
    import stat
    os.chmod(cr_filename, stat.S_IREAD|stat.S_IWRITE)


  def _open_account(self, new_username, new_password, test_account):
    """called when the creadentials change - is responsible for cleaning up the UI,
    changing the ds_client and optionally loading stored messages for the new
    account"""
    if self.username == new_username and self.test_account == test_account:
      if self.password == new_password:
        return # we do not need to do anything, everything is the same
    else:
      # we need to replace the content of the UI
      shelf = self._open_shelf()
      if self.username:
        # we have a username, so we save the current state
        self._store_to_shelf(shelf)
      self.received_messages = {}
      self.sent_messages = {}
      self.message_data = {}
      self.account_info = None
      self.username = new_username
      self.test_account = test_account
      self.sent_message_store.clear()
      self.received_message_store.clear()
      self.received_attachment_store.clear()
      self.received_message_textview.get_buffer().set_text("")
      self.sent_message_textview.get_buffer().set_text("")
      self.builder.get_object("account_info_textbuffer").set_text("")
      self.builder.get_object("download_complete_message").set_sensitive(False)
      self._load_from_shelf(shelf)
      shelf.close()
    self.password = new_password
    self._check_zero_accounts()
    self._check_empty_message_stores()

  def _open_known_account(self, username):
    for i, row in enumerate(self.account_store):
      if row[0] == username:
        self.account_treeview.set_cursor(i)
        break

  def _open_shelf(self):
    import shelve
    fname = os.path.join(GUI.get_dsgui_private_dir(create=True), "messages.shelf")
    encoding = sys.getfilesystemencoding()
    if encoding:
       fname = fname.encode(encoding)
    shelf = shelve.open(fname, protocol=-1)
    return shelf


  def _store_to_shelf(self, shelf):
    key = self._get_shelf_key()
    shelf[key] = {"received_messages":self.received_messages,
                  "sent_messages":self.sent_messages,
                  "message_data":self.message_data,
                  "account_info":self.account_info,
                  }
    shelf.sync()


  def _load_from_shelf(self, shelf):
    data = shelf.get(self._get_shelf_key(), None)
    if data:
      message_data = data.get("message_data", None)
      if message_data:
        self.message_data = message_data
      for id,message in data['received_messages'].iteritems():
        self.add_received_message(message)
      for id,message in data['sent_messages'].iteritems():
        self.add_sent_message(message)
      account_info = data.get("account_info", None)
      self.set_account_info(account_info)

  def _get_shelf_key(self):
    """returns a key to be used for shelf based on current credentials"""
    return GUI._create_shelf_key(self.username, self.test_account)


  def _show_error(self, message, exception=None):
    if exception:
      message += "\n\n"+_("Details:")+" "+str(exception)
      exc_type, exc_value, tb = sys.exc_info()
      import traceback
      tb_lines = traceback.format_tb(tb)
      if len(tb_lines) > 5:
        # limit traceback to 20 lines
        tb_lines = ["..."]+tb_lines[-5:]
      traceback_text = "\n".join(tb_lines)
      message += "\n\n"+traceback_text
    d = gtk.MessageDialog(self.window1,
                          message_format=message,
                          buttons=gtk.BUTTONS_OK,
                          type=gtk.MESSAGE_ERROR)
    d.run()
    d.destroy()

  def _handle_authorization_error(self, e, message_header=None):
    """convenience method to display information about failed authorization;
    returns True if the exception was handled, False otherwise"""
    if type(e.args) in (tuple,list) and \
    len(e.args) > 0 and type(e.args[0]) == tuple and len(e.args[0]) > 0:
      if e.args[0][0] == 401:
        # Unauthorized
        message = message_header and message_header+"\n\n" or ""
        message += _("""Authorization failed!

Please check your credentials including the test-environment setting.
It might be necessary to log into the DS system using your browser and change your password.""")
        self._show_error(message)
        return True
    return False
  
  def _save_credentials(self, config):
    """save current creadentials to a file"""
    for i, row in enumerate(self.account_store):
      num = ((i>0) and str(i+1) or "")
      section = "credentials"+num
      config.add_section(section)
      username, password, test_account = list(row)[0:3]
      config.set(section, 'username', username)
      config.set(section, 'password', password)
      config.set(section, 'test_account', test_account)
    current = self._get_selected_account()
    if current:
      config.add_section("default_account")
      config.set("default_account", "username", current[0])

  def _save_gui_state(self, config):
    config.add_section("message_ordering")
    for tree_name in ("received", "sent"):
      rec_col, rec_ord = getattr(self, tree_name+"_message_store").get_sort_column_id()
      if rec_col:
        config.set('message_ordering', tree_name+'_sort_column', rec_col)
        if rec_ord != None:
          rec_ord = rec_ord.value_name[4:]
          config.set('message_ordering', tree_name+'_sort_order', rec_ord)

  def _save_connection_state(self, config):
    config.add_section("connection")
    config.set('connection', 'https_proxy', self.proxy)

  def _load_gui_state(self, config):
    for tree_name in ("received", "sent"):
      if config.has_option("message_ordering", tree_name+"_sort_column"):
        rec_col = config.getint("message_ordering", tree_name+"_sort_column")
        rec_ord = gtk.SORT_ASCENDING
        if config.has_option("message_ordering", tree_name+"_sort_order"):
          try:
            rec_ord = getattr(gtk, config.get("message_ordering", tree_name+"_sort_order"))
          except:
            pass
        getattr(self, tree_name+"_message_store").set_sort_column_id(rec_col, rec_ord)
  
  def _load_connection_state(self, config):
    if not self.proxy_support_ok:
      # this version of Python does not support HTTPS proxy anyway
      self.proxy = None
      return
    if config.has_section("connection"):
      if config.has_option('connection', 'https_proxy'):
        proxy = config.get('connection', 'https_proxy')
        if proxy == "-1":
          self.proxy = -1
        elif proxy == "None":
          self.proxy = None
        else:
          self.proxy = proxy
            
  def _load_config_file(self):
    cr_filename = self._get_config_filename()
    import ConfigParser
    config = ConfigParser.RawConfigParser()
    config.read(cr_filename)
    i = 1
    section = "credentials" + ((i>1) and str(i) or "")
    while config.has_section(section):
      if config.has_option(section, "username"):
        username = config.get(section, "username")
      else:
        username = ""
      if config.has_option(section, "password"):
        password = config.get(section, "password")
      else:
        password = ""
      if config.has_option(section, "test_account"):
        test_account = config.getboolean(section, "test_account")
      else:
        test_account = True
      self.add_account(username, password, test_account)
      i += 1
      section = "credentials" + ((i>1) and str(i) or "")
    self._load_gui_state(config)
    self._load_connection_state(config)
    if config.has_option("default_account", "username"):
      username = config.get("default_account", "username")
      self._open_known_account(username)
    elif len(self.account_store):
      self.account_treeview.set_cursor(0)
    #self._open_account(username, password, test_account)


  def _get_config_filename(self):
    return os.path.join(GUI.get_dsgui_private_dir(create=True), "dsgui.conf")
  
  def _check_empty_message_stores(self):
    """check if the store of received or sent messages is empty
    and warn if need be.
    """
    for treeview,message in ((self.received_message_treeview,[0,_("No received messages"),"","","",True]),
                             (self.sent_message_treeview,[0,_("No sent messages"),"","","","",""])):
      model = treeview.get_model()
      for i,row in enumerate(model):
        if row[0] == 0:
          model.remove(model.get_iter(i))
          break
      if len(model) == 0:
        model.append(message)
        treeview.set_sensitive(False)
      else:
        treeview.set_sensitive(True)

  def _activate_tooltips_for_main_menu(self):
    def _X_(w):
      if w.get_action():
        tooltip = w.get_action().get_property("tooltip")
        if tooltip:
          self.statusbar.push(-1, tooltip)
    def _Y_(w):
      self.statusbar.pop(-1)
    def _act_one(x):
      for child in x:
        child.connect("select", _X_)
        child.connect("deselect", _Y_)
        submenu = child.get_submenu()
        if submenu:
          _act_one(submenu)
    menu = self.main_menu
    _act_one(menu)
    
  def _post_builder_ui_compatibility_fixes(self, gui_xml):
    if gtk.gtk_version < (2,16,0):
      # in 2.14 menu is not properly loaded from XML
      import xml.dom.minidom as dom
      doc = dom.parseString(gui_xml)
      from ui_xml_versions import manually_create_menu, bind_nonmenu_action_references
      menu = manually_create_menu(doc, "main_menu", self, self.builder)
      bind_nonmenu_action_references(doc, self.builder)
      old_menu = self.builder.get_object("main_menu")
      if old_menu:
        old_menu.destroy()
      self.builder.get_object("vbox1").pack_start(menu, False, False)
      self.builder.get_object("vbox1").reorder_child(menu, 0)
      menu.show_all()
    else:
      menu = self.builder.get_object("main_menu")
    self.main_menu = menu
    # now try hildon support (Maemo)
    try:
      import hildon
    except:
      pass
    else:
      self.app = hildon.Program()
      hildon_window = hildon.Window()
      self.app.add_window(hildon_window)
      self.builder.get_object("vbox1").reparent(hildon_window)
      hildon_window.set_title("dsgui")
      self.window1.destroy()
      new_menu = gtk.Menu()
      for child in menu.get_children():
        child.reparent(new_menu)
      hildon_window.set_menu(new_menu)
      menu.destroy()
      self.builder.get_object("toolbar1").destroy()
      self.builder.get_object("warning_label").destroy()
      hildon_window.connect("destroy", self.on_window1_destroy)
      self.window1 = hildon_window
    
  def try_client_method(self, method, args, error_message=None, handle_status=True):
    """convenience function to run a method of self.client and display
    error messages, etc. as appropriate.
    Returns a tuple (reply, error_shown), where reply is the reply of the method and
    error_shown is a boolean telling if an error message was displayed.
    """
    if not error_message:
      error_message = _("An error occured during the following operation: %s") % method.__name__
    try:
      reply = method(*args)
    except Exception, e:
      if not self._handle_authorization_error(e, message_header=error_message):
        #print e
        self._show_error(error_message, exception=e)
      return None, True
    else:
      error_shown = False
      if handle_status:
        code, message = GUI._extract_code_and_message_from_reply(reply)
        if code != "0000":
          error_message += "\n\n"
          error_message += _("Reply status: %(message)s\nReply code: %(code)s")%{"message":message,"code":code}
          self._show_error(error_message)
          error_shown = True
      return reply, error_shown
    
  @classmethod
  def get_dsgui_private_dir(cls, create=False):
    homedir = cls.get_home_dir()
    priv_dir = os.path.join(homedir, ".dsgui")
    if create and not os.path.exists(priv_dir):
      os.mkdir(priv_dir)
    return priv_dir
  
  @classmethod
  def get_home_dir(cls):
    dir = None
    if os.name == "nt":
      try:
        import _winreg as reg
        dir = reg.QueryValueEx(reg.OpenKey(reg.HKEY_CURRENT_USER, "Volatile Environment"), "APPDATA")[0]
      except EnvironmentError:
        pass
    if not dir:
      dir = os.getenv('HOME') or cls.topdir
    return dir
    
  @classmethod
  def _get_account_type_as_text(self, test_account):
    account_type = test_account and _("Test account") or _("Standard account")
    return account_type
      
  @classmethod
  def _extract_code_and_message_from_reply(cls, reply):
    if not reply:
      return None, None
    if not hasattr(reply, 'status'):
      return None, None
    elif hasattr(reply.status, 'dmStatusCode'):
      return reply.status.dmStatusCode, reply.status.dmStatusMessage
    elif hasattr(reply.status, 'dbStatusCode'):
      return reply.status.dbStatusCode, reply.status.dbStatusMessage
    return None, None

  @classmethod
  def _create_ds_client(self, username, password, test_account, proxy=None):
    return Client(username, password, test_environment=test_account, proxy=proxy)


  @classmethod
  def _create_temp_file_for_dmfile(self, dmfile):
    import tempfile
    ext = os.path.splitext(dmfile._dmFileDescr)[1]
    temp = tempfile.mkstemp(suffix=ext)
    os.write(temp[0], dmfile.get_decoded_content())
    os.close(temp[0])
    return temp[1]

  @classmethod
  def _format_datetime(cls, d):
    if d:
      return d.strftime(cls.date_time_format)
    return ""

  @classmethod
  def _create_shelf_key(cls, username, test_account):
    """returns a key to be used for shelf based on current credentials"""
    return "%s___%s" % (username, test_account)

  @classmethod
  def fix_ui_xml(cls, filename):
    """convenience function wrapping several one-purpose methods"""
    gui_xml = cls.compatibilize_ui_xml(filename).encode('utf-8')
    gui_xml = cls.translate_ui_xml(gui_xml)
    return gui_xml
  
  @classmethod
  def translate_ui_xml(cls, gui_xml):
    gui_xml = translate_xml_string(gui_xml, _)
    return gui_xml

  @classmethod
  def compatibilize_ui_xml(cls, filename):
    from ui_xml_versions import compatibilize_xml_file
    xml_string = compatibilize_xml_file(os.path.join(cls.topdir,filename), gtk.gtk_version)
    return xml_string

  def _check_zero_accounts(self):
    actions = ("connect_action", "find_data_box", "send_message")
    if self.username:
        self.client = self._create_ds_client(self.username, self.password, self.test_account, proxy=self.proxy)
        for action in actions:
            self.builder.get_object(action).set_sensitive(True)
    else:
        self.client = None
        for action in actions:
            self.builder.get_object(action).set_sensitive(False)



if __name__ == "__main__":
  import logging
  logging.basicConfig(level=logging.CRITICAL)
##   logging.getLogger('suds.client').setLevel(logging.DEBUG)
  app = GUI()
  gtk.main()

